// FILE: land.cpp
// Written by: Grant Macklem (Grant.Macklem@colorado.edu)
// Last modified: May 1, 2002

// Defines
#define DEBUG

// Includes
#include <stdlib.h>             // Provides EXIT_SUCCESS
#include <GL/glut.h>            // Provides OpenGL commands
#include <iostream>             // Provides cout, cin
#include <fstream>
#include <iomanip>
//#include <math.h>               // Provides log
#include <float.h>              // Provides DBL_MAX, DBL_MIN
#include "land.h"               // Provides prototypes, etc.

using namespace std;

#ifdef DEBUG
    #include <assert.h>
#endif

void read_landscape( );
double random_real( double low, double high );
double random_fraction( );
double min( double a, double b );
void calc_min_max_height( );
void calc_min_max_height( GLdouble** array, int limit );
void draw_middle_i( int i );
void draw_middle_j( int j );


// Constants, see header file
const GLint GON_SIZE = 16;                   // Polygon size
const GLint FIELD_SIZE = 1024;               // Square landscape, power of two
const GLint LIMIT = FIELD_SIZE / GON_SIZE;   // Must be an integer
//const GLdouble HEIGHT_MAX = 100.25;     // Limiting height for random landscape
const GLdouble HEIGHT_MAX = 400;             // Limiting height for random landscape
const GLdouble EPSILON = HEIGHT_MAX / 10;    // Epsilon for thresholding

// Global variables, in header file for global scope
GLdouble** grid;
GLdouble** patchgrid;
//GLdouble grid[LIMIT+1][LIMIT+1];
GLdouble* points;
//GLint base[LIMIT + 1];
// The interp 2d array holds twice as many points as grid, for linear
// interpolation.
GLdouble** interp;
GLint* base;
GLint used = 0;
GLfloat xrot = -60.0, yrot = -10.0, zrot = -25.0;
double hmin, hmax;

unsigned int seed;
GLdouble zoom = 1;
// This is for maximal level.  Set to an integer for intermediate results.
int LEVEL = LIMIT;
//int LEVEL = 32;

int main(int argc, char** argv)
{
    int j;

    // Initialize dynamic memory
    grid = new GLdouble*[LIMIT+1];             // Set up the rows
    for (j = 0; j <= LIMIT; j++)
        grid[j] = new GLdouble[LIMIT+1];       // Set up the columns
    patchgrid = new GLdouble*[LIMIT+1];        // Set up the rows
    for (j = 0; j <= LIMIT; j++)
        patchgrid[j] = new GLdouble[LIMIT+1];  // Set up the columns
    interp = new GLdouble*[2*LIMIT+1];         // Set up the rows
    for (j = 0; j <= 2*LIMIT; j++)
        interp[j] = new GLdouble[2*LIMIT+1];   // Set up the columns

 /*****************************************************************************
    Every internal vertex has its value stored 4 times.
    Where it multiples by (limit-1) because limit is the number of squares,
    meaning there are limit + 1 edges.  Thus, to cut off both edges and get
    the number of internal edges is (limit+1) - 2 = limit - 1
        = 4*(limit - 1) * (limit - 1) = 4*limit^2 -8*limit +4
    Every internal edge vertex (everything but the corners) has its data
    stored twice (2 times the length of each of 4 sides)
      = 2 * (limit - 1) * 4 = 8*limit - 8
    Every corner is stored once
      = 4
    sum = 4 * limit^2 - 8*limit + 4 + 8*limit - 8 +4
          = 4 * limit^2
 *****************************************************************************/

    points = new GLdouble[LIMIT * LIMIT * 4];
    base = new GLint[LIMIT + 1];

    cout << "Random Landscape Generation and Smoothing\n";
    cout << "University of Colorado at Boulder\n";
    cout << "Department of Computer Science\n";
    cout << "Grant Macklem\n";
    cout << "May, 2003\n\n";

    // Seed the random number generator
    cout << "Enter a number to seed the random number generator: ";
    cin >> seed;
    srand(seed);
    cout << "Generating Landscape\n";
    //read_landscape( );
    //calc_min_max_height( grid, LIMIT/2 );
    generate_landscape(0, FIELD_SIZE, FIELD_SIZE, 0, 0, 0, 0, 0);
    fix_matrix();
    used = 0;


    cout << "\nUse the following commands to work with the landscape\n";
    cout << "(the graphic window must be active)\n";
    cout << "/\tTurn counterclockwise\n";
    cout << "\\\tTurn clockwise\n";
    cout << "(up)\tFlip toward screen\n";
    cout << "(down)\tFlip away from screen\n";
    cout << "(left)\tFlip left\n";
    cout << "(right)\tFlip right\n";
    cout << "r\tRegenerate the landscape if randomly generated\n";
    cout << "l\tSmooth with Lagrange interpolating polynomial\n";
    cout << "c\tSmooth with Bezier curves\n";
    cout << "p\tSmooth with Bezier patches\n";
    cout << "m\tView the Bezier patch smoothed landscape\n";
    cout << ",\tView the Bezier curve smoothed landscape\n";
    cout << ".\tView the Lagrange smoothed landscape\n";
    cout << "q\tQuit the program" << endl;

    // Initialize the windowing environment
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(600, 600);
    glutInitWindowPosition(100,100);
    glutCreateWindow("Landscape");
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutSpecialFunc(special);
    glutKeyboardFunc(keyboard);

    init();
    glutMainLoop();

    return EXIT_SUCCESS;
}


void init()
{
    glClearColor(1.0, 1.0, 1.0, 0.0);
    glClearDepth(1.0);
    glShadeModel(GL_SMOOTH);
    glEnable (GL_DEPTH_TEST);
    glLineWidth(2.0);
}


void read_landscape( )
{
    ifstream ins;
    //ins.open("bosworthclouds.txt");
    ins.open("daubechies");
    for ( int i = 0; i <= LIMIT; i++ )
    {
        for ( int j = 0; j <= LIMIT; j++ )
        {
            ins >> grid[i][j];
        }
    }
    ins.close();
}


void write_out( )
{
    ofstream outs;
    outs.open( "heights.txt" );
    for( int i = 0; i <= LIMIT; i++ )
    {
        for ( int j = 0; j <= LIMIT; j++ )
        {
            outs << setprecision(10) << grid[i][j] << " ";
        }
        outs << endl;
    }
    outs.close();
}


void generate_landscape(GLdouble left, GLdouble right, GLdouble top,
    GLdouble bottom, GLdouble top_left_height, GLdouble top_right_height,
    GLdouble bottom_left_height, GLdouble bottom_right_height)
{

#ifdef DEBUG
    assert (left < right);
    assert (bottom < top);
#endif


    GLdouble width = right - left;
    GLdouble new_height, x_middle, y_middle;

    if(width < GON_SIZE + 1)
    {   // Store the current points
        points[used++] = top_left_height;
        points[used++] = top_right_height;
        points[used++] = bottom_left_height;
        points[used++] = bottom_right_height;
    }
    else
    {   // Subdivide and recurse on four subrectangles

        // Calculate new height between -HEIGHT_MAX and +HEIGHT_MAX
        new_height = ( top_left_height + top_right_height + bottom_left_height
            + bottom_right_height ) / 4.0;
        new_height += random_real( -min( width, HEIGHT_MAX ), min( width, HEIGHT_MAX ) );
        //new_height = HEIGHT_MAX + -2 * HEIGHT_MAX * rand() / (RAND_MAX + 1.0);

        x_middle = (left + right) / 2.0;
        y_middle = (top + bottom) / 2.0;

        // Recurse on upper left corner
        generate_landscape(left, x_middle, top, y_middle,
        top_left_height,
        (top_left_height + top_right_height) / 2.0,
        (top_left_height + bottom_left_height) / 2.0,
        new_height);

        // Recurse on upper right corner
        generate_landscape(x_middle, right, top, y_middle,
        (top_left_height + top_right_height) / 2.0,
        top_right_height,
        new_height,
        (top_right_height + bottom_right_height) / 2.0);

        // Recurse on lower left corner
        generate_landscape(left, x_middle, y_middle, bottom,
        (top_left_height + bottom_left_height) / 2.0,
        new_height,
        bottom_left_height,
        (bottom_left_height + bottom_right_height) / 2.0);

        // Recurse on lower right corner
        generate_landscape(x_middle, right, y_middle, bottom,
        new_height,
        (top_right_height + bottom_right_height) / 2.0,
        (bottom_left_height + bottom_right_height) / 2.0,
        bottom_right_height);
    }
}


double random_fraction( )
{
    return rand( ) / double( RAND_MAX );
}

double random_real( double low, double high )
{
    return low + random_fraction( ) * ( high - low );
}

double min( double a, double b )
{
    if ( a < b )
        return a;
    else
        return b;
}


void fix_matrix()
{
    int i, j;
    calc_min_max_height( );

    // When i is even, base is 4 * base[i/2].  When i is odd, base is
    // 4 + 4 * base[i/2].  Thus, this statement works
    base[0] = 0;

    // Probably just needs less than since the final point is not referenced
    // as being the final point
    for(i = 0; i < LIMIT; i++)
    base[i] = 4 * base[i/2] + (i % 2) * 4;

    grid[0][0] = points[0];

    // For all points except final column and final row
    for(i = 0; i < LIMIT; i++)
    {
        for(j = 0; j < LIMIT; j++)
        {
            grid[i][j] = points[2 * base[i] + base[j]];
        }
    }

    // Do final row and column, except for very last point
    for (j = 0; j < LIMIT; j++)
    {
        grid[LIMIT][j] = points[2 * base[LIMIT - 1] + 2 + base[j]];
        grid[j][LIMIT] = points[2 * base[j] + base[LIMIT - 1] + 1];
    }

    // Do final point
    // Really [2 * base[LIMIT - 1] + 3 + base[LIMIT - 1]]
    grid[LIMIT][LIMIT] = points[3 * base[LIMIT - 1] + 3];

}

/*
void display()
{
    static int i, j, k, m;
    GLdouble red, green, blue;      // For coloring

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslatef(0, 0, -.8*FIELD_SIZE);
    glRotatef(clamp(xrot), 1.0, 0.0, 0.0);
    glRotatef(clamp(yrot), 0.0, 1.0, 0.0);
    glRotatef(clamp(zrot), 0.0, 0.0, 1.0);


    //glBegin(GL_TRIANGLES);           // Comment out to draw lines
    for (i = -LIMIT/2; i < LIMIT/2; i++)
    {
        for (j = -LIMIT/2; j < LIMIT/2; j++)
        {
            k = i + LIMIT/2;
            m = j + LIMIT/2;
            glBegin(GL_TRIANGLES);      // Only for drawing lines (uncommented)
            get_rgb(grid[k][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);         // 0

            get_rgb(grid[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, j*GON_SIZE, grid[k+1][m]);       // 2

            get_rgb(grid[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+1)*GON_SIZE, grid[k][m+1]);     // 1



            get_rgb(grid[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, j*GON_SIZE, grid[k+1][m]);       // 2

            get_rgb(grid[k+1][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, (j+1)*GON_SIZE, grid[k+1][m+1]); // 3

            get_rgb(grid[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+1)*GON_SIZE, grid[k][m+1]);     // 1

            ///*       // Only for drawing lines to accentuate the landscape (uncommented)
            glEnd();

            glBegin(GL_LINES);
            glColor3f(0,0,0);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);
            glVertex3f((i+1) * GON_SIZE, j * GON_SIZE, grid[k+1][m]);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);
            glVertex3f(i * GON_SIZE, (j+1) * GON_SIZE, grid[k][m+1]);
            glEnd();
            //*/
/*
        }
    }
    //glEnd();             // Comment out to draw lines
    glutSwapBuffers();
}
*/


void display()
{
    static int i, j, k, m;
    GLdouble red, green, blue;      // For coloring

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslatef(0, 0, -.8*FIELD_SIZE);
    glRotatef(clamp(xrot), 1.0, 0.0, 0.0);
    glRotatef(clamp(yrot), 0.0, 1.0, 0.0);
    glRotatef(clamp(zrot), 0.0, 0.0, 1.0);


    //glBegin(GL_TRIANGLES);           // Comment out to draw lines
    int step = LIMIT / LEVEL;
    for (i = -LIMIT/2; i < LIMIT/2; i+=step)  // i++
    {
        for (j = -LIMIT/2; j < LIMIT/2; j+=step) //j++
        {
            k = i + LIMIT/2;
            m = j + LIMIT/2;
            glBegin(GL_TRIANGLES);      // Only for drawing lines (uncommented)
            get_rgb(grid[k][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);         // 0

            get_rgb(grid[k+step][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+step)*GON_SIZE, j*GON_SIZE, grid[k+step][m]);       // 2

            get_rgb(grid[k][m+step], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+step)*GON_SIZE, grid[k][m+step]);     // 1



            get_rgb(grid[k+step][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+step)*GON_SIZE, j*GON_SIZE, grid[k+step][m]);       // 2

            get_rgb(grid[k+step][m+step], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+step)*GON_SIZE, (j+step)*GON_SIZE, grid[k+step][m+step]); // 3

            get_rgb(grid[k][m+step], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+step)*GON_SIZE, grid[k][m+step]);     // 1

            ///*       // Only for drawing lines to accentuate the landscape (uncommented)
            glEnd();

            glBegin(GL_LINES);
            glColor3f(0,0,0);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);
            glVertex3f((i+step) * GON_SIZE, j * GON_SIZE, grid[k+step][m]);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);
            glVertex3f(i * GON_SIZE, (j+step) * GON_SIZE, grid[k][m+step]);
            glEnd();
            //*/
        }
    }
/*
    draw_middle_i( -LIMIT/2 );
    draw_middle_i( LIMIT/2 );
    draw_middle_j( -LIMIT/2 );
    draw_middle_j( LIMIT/2 );
*/
    //glEnd();             // Comment out to draw lines
    glutSwapBuffers();
}


void draw_middle_i( int i )
{
    int j, k, m;
    GLdouble red, green, blue;      // For coloring
    int step = LIMIT / LEVEL;


    for (j = -LIMIT/2; j < LIMIT/2; j+=step) //j++
    {
        k = i + LIMIT/2;
        m = j + LIMIT/2;
        glBegin(GL_QUADS);
        get_rgb(grid[k][m], red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);         // 0

        get_rgb(grid[k][m+step], red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, (j+step)*GON_SIZE, grid[k][m+step]);     // 1

        get_rgb(hmin, red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, (j+step)*GON_SIZE, hmin);       // 2

        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, j*GON_SIZE, hmin);       // 2
        glEnd();
    }
}


void draw_middle_j( int j )
{
    int i, k, m;
    GLdouble red, green, blue;      // For coloring
    int step = LIMIT / LEVEL;


    for (i = -LIMIT/2; i < LIMIT/2; i+=step) //j++
    {
        k = i + LIMIT/2;
        m = j + LIMIT/2;
        glBegin(GL_QUADS);
        get_rgb(grid[k][m], red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, j * GON_SIZE, grid[k][m]);         // 0

        get_rgb(grid[k+step][m], red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f((i+step) * GON_SIZE, j * GON_SIZE, grid[k+step][m]);     // 1

        get_rgb(hmin, red, green, blue);
        glColor3f(red, green, blue);
        glVertex3f((i+step)*GON_SIZE, j * GON_SIZE, hmin);       // 2

        glColor3f(red, green, blue);
        glVertex3f(i * GON_SIZE, j*GON_SIZE, hmin);       // 2
        glEnd();
    }
}


void calc_min_max_height( )
{
    hmin = DBL_MAX;
    hmax = DBL_MIN;
    for ( int i = 0; i < used; i++ )
    {
        if ( points[i] < hmin )
            hmin = points[i];
        if ( points[i] > hmax )
            hmax = points[i];
    }
}



void calc_min_max_height( GLdouble** array, int limit )
{
    hmin = DBL_MAX;
    hmax = DBL_MIN;
    int i, j, k, m;


    for (i = -limit; i < limit; i++)  // i++
    {
        for (j = -limit; j < limit; j++) //j++
        {
            k = i + limit;
            m = j + limit;

            if ( array[k][m] < hmin )
                hmin = array[k][m];
            if ( array[k][m] > hmax )
                hmax = array[k][m];
        }
    }
}


void display_lagrange()
{
    static int i, j, k, m;

    GLint gon_size = GON_SIZE / 2;  // Need half the gon size, since points are
                                    // interpolated inbetween
    GLdouble red, green, blue;      // For coloring

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslatef(0, 0, -.8*FIELD_SIZE);
    glRotatef(clamp(xrot), 1.0, 0.0, 0.0);
    glRotatef(clamp(yrot), 0.0, 1.0, 0.0);
    glRotatef(clamp(zrot), 0.0, 0.0, 1.0);

    //glBegin(GL_TRIANGLES);           // Comment out to draw lines
    for (i = -LIMIT; i < LIMIT; i++)
    {
        for (j = -LIMIT; j < LIMIT; j++)
        {
            k = i + LIMIT;
            m = j + LIMIT;
            glBegin(GL_TRIANGLES);      // Only for drawing lines (uncommented)

            get_rgb(interp[k][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * gon_size, j * gon_size, interp[k][m]);         // 0

            get_rgb(interp[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*gon_size, j*gon_size, interp[k+1][m]);       // 2

            get_rgb(interp[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * gon_size, (j+1)*gon_size, interp[k][m+1]);     // 1
            


            get_rgb(interp[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*gon_size, j*gon_size, interp[k+1][m]);       // 2

            get_rgb(interp[k+1][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*gon_size, (j+1)*gon_size, interp[k+1][m+1]); // 3

            get_rgb(interp[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * gon_size, (j+1)*gon_size, interp[k][m+1]);     // 1

            ///* // Only for drawing lines to accentuate the landscape (uncommented)
            glEnd();

            glBegin(GL_LINES);
            glColor3f(0,0,0);
            glVertex3f(i * gon_size, j * gon_size, interp[k][m]);
            glVertex3f((i+1) * gon_size, j * gon_size, interp[k+1][m]);
            glVertex3f(i * gon_size, j * gon_size, interp[k][m]);
            glVertex3f(i * gon_size, (j+1) * gon_size, interp[k][m+1]);
            glEnd();
            //*/
        }
    }
    //glEnd();             // Comment out to draw lines
    glutSwapBuffers();
}


void display_patch()
{
    static int i, j, k, m;
    GLdouble red, green, blue;      // For coloring

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslatef(0, 0, -.8*FIELD_SIZE);
    glRotatef(clamp(xrot), 1.0, 0.0, 0.0);
    glRotatef(clamp(yrot), 0.0, 1.0, 0.0);
    glRotatef(clamp(zrot), 0.0, 0.0, 1.0);


    //glBegin(GL_TRIANGLES);           // Comment out to draw lines
    for (i = -LIMIT/2; i < LIMIT/2; i++)
    {
        for (j = -LIMIT/2; j < LIMIT/2; j++)
        {
            k = i + LIMIT/2;
            m = j + LIMIT/2;
            glBegin(GL_TRIANGLES);      // Only for drawing lines (uncommented)
            get_rgb(patchgrid[k][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, patchgrid[k][m]);         // 0

            get_rgb(patchgrid[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, j*GON_SIZE, patchgrid[k+1][m]);       // 2

            get_rgb(patchgrid[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+1)*GON_SIZE, patchgrid[k][m+1]);     // 1



            get_rgb(patchgrid[k+1][m], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, j*GON_SIZE, patchgrid[k+1][m]);       // 2

            get_rgb(patchgrid[k+1][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f((i+1)*GON_SIZE, (j+1)*GON_SIZE, patchgrid[k+1][m+1]); // 3

            get_rgb(patchgrid[k][m+1], red, green, blue);
            glColor3f(red, green, blue);
            glVertex3f(i * GON_SIZE, (j+1)*GON_SIZE, patchgrid[k][m+1]);     // 1

            ///*       // Only for drawing lines to accentuate the landscape (uncommented)
            glEnd();

            glBegin(GL_LINES);
            glColor3f(0,0,0);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, patchgrid[k][m]);
            glVertex3f((i+1) * GON_SIZE, j * GON_SIZE, patchgrid[k+1][m]);
            glVertex3f(i * GON_SIZE, j * GON_SIZE, patchgrid[k][m]);
            glVertex3f(i * GON_SIZE, (j+1) * GON_SIZE, patchgrid[k][m+1]);
            glEnd();
            //*/
        }
    }
    //glEnd();             // Comment out to draw lines
    glutSwapBuffers();
}


// This function provides a linear interpolation between high_color and
// low_color, both specified in their RGB components (0-255)
void get_rgb(GLdouble x, GLdouble& red, GLdouble& green, GLdouble& blue)
{
    int high_color[3] = {0, 255, 0};
    int low_color[3] = {0, 0, 255};
    //int high_color[3] = {255, 0, 0};
    //int low_color[3] = {154, 84, 34};

    // Dark Green   {0, 139, 0};
    // Spring Green {0, 238, 118};

    // Sandy Brown  {244, 164, 96};
    // Saddle Brown {139, 69, 19};
/*
    red   = ((double)(high_color[0] - low_color[0])) / (2*HEIGHT_MAX)
            * (x - HEIGHT_MAX) + high_color[0];
    green = ((double)(high_color[1] - low_color[1])) / (2*HEIGHT_MAX)
            * (x - HEIGHT_MAX) + high_color[1];
    blue  = ((double)(high_color[2] - low_color[2])) / (2*HEIGHT_MAX)
            * (x - HEIGHT_MAX) + high_color[2];
*/

    // This will calculate between the max and min actual heights
    // instead of the maximum allowed height

    red   = ((double)(high_color[0] - low_color[0])) / (hmax - hmin) 
        * (x - hmin) + low_color[0];
    green = ((double)(high_color[1] - low_color[1])) / (hmax - hmin) 
        * (x - hmin) + low_color[1];
    blue  = ((double)(high_color[2] - low_color[2])) / (hmax - hmin)
        * (x - hmin) + low_color[2];

    // Scale to a fraction from 0 to 1
    red   /= 255;
    green /= 255;
    blue  /= 255;

}


void special(int key, int x, int y)
{
    switch(key)
    {
        case GLUT_KEY_UP:
            xrot += 5.0;
            break;
        case GLUT_KEY_DOWN:
            xrot -= 5.0;
            break;
        case GLUT_KEY_LEFT:
            yrot -= 5.0;
            break;
        case GLUT_KEY_RIGHT:
            yrot += 5.0;
            break;
        default:
            return;
    }

    glutPostRedisplay();
}


void reshape(int w, int h)
{
    double factor = 0.7;
    glViewport(0, 0, (GLint) w, (GLint) h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(zoom * -factor*FIELD_SIZE, zoom * factor*FIELD_SIZE, zoom * -factor*FIELD_SIZE,
    zoom * factor*FIELD_SIZE, 5.0, 1.74*FIELD_SIZE);
}


inline GLfloat clamp(GLfloat& angle)
{
    /* Must use greater than or equal to because on systems that only have
    * software rendering available, the keyboard buffer may get filled with
    * commands to rotate while the display is rendering the last rotated frame
    * thereby increasing the angle past 360 degrees.
    */

    // Probably faster than using abs
    if ( (angle >= 360.0) || (angle <= -360.0) )
        angle = 0;
    return angle;
}


void keyboard(unsigned char key, int x, int y)
{
    int i;
    switch (key)
    {
        case '/':
            zrot += 5.0;
            break;
        case '\\':
            zrot -= 5.0;
            break;
        case 'q': case 'Q': case (char) 27: /* Escape */
        // Free dynamic memory, C++ style
            for (i = 0; i <= LIMIT;  i++)
            {
                delete[] grid[i];            // STEP 1: DELETE THE COLUMNS
                delete[] patchgrid[i];
            }
            delete[] grid;                   // STEP 2: DELETE THE ROWS
            delete[] patchgrid;
            for (i = 0; i < 2*LIMIT; i++)
                delete[] interp[i];          // Delete the columns
            delete[] interp;                 // Delete the rows
            delete[] points;
            delete[] base;
            ::exit(0);
            break;
      /*case 'p':     // print rotations, debug only
            cout << "xrot: " << xrot << "\nyrot: " << yrot << "\nzrot: " << zrot
                 << endl;
            break; */
      case 'w':     // Write out the height map to disk
          write_out( );
          break;
      case 'l':     // Smooth with Lagrange polynomial
            smooth_lagrange();
            cout << "Smoothed with Lagrange." << endl;
            glutDisplayFunc(display_lagrange);
            calc_min_max_height( interp, LIMIT );
            break;
      case 'c':     // Smooth with Bzier curves
            smooth_curve();
            cout << "Smoothed with Bezier curve." << endl;
            glutDisplayFunc(display);
            calc_min_max_height( grid, LIMIT/2 );
            break;
      case 'p':     // Smooth with Bzier patches
            smooth_patch();
            cout << "Smoothed with Bezier Patch." << endl;
            calc_min_max_height( patchgrid, LIMIT/2 );
            glutDisplayFunc(display_patch);
            break;
      case 'r':     // Regenerate
            srand(seed);
            generate_landscape(0, FIELD_SIZE, FIELD_SIZE, 0, 0, 0, 0, 0);
            fix_matrix();
            used = 0;
            glutDisplayFunc(display);
            cout << "Landscape regenerated." << endl;
            break;
      case 'z':     // Debug only, zoom in
            zoom -= 0.01;
            break;
      case 'a':     // Debug only, zoom out
            zoom += 0.01;
            break;
      case '1':     // Debug only, refresh view to show zooms
            reshape(600,600);
            break;
      case ',':  // Switch to Bezier Curve view
            glutDisplayFunc(display);
            break;
      case '.':  // Switch to Lagrange Poly view
            glutDisplayFunc(display_lagrange);
            break;
      case 'm':  // Switch to Bezier Patch view
            glutDisplayFunc(display_patch);
            break;
      default:
            return;
    }

    glutPostRedisplay();
}
