// FILE: smooth.cpp
// Written by: Grant Macklem (Grant.Macklem@colorado.edu)
// Last modified: May 1, 2002

#include "land.h"
#include <string.h>   // For memcpy

// This routine smooths the landscape with a Bezier curve.
void smooth_curve()
{
    int i, j, endpoint;
    GLdouble p0, p1, p2, p3, t;

    endpoint = LIMIT - 3;

//  Do the rows
    for (i = 0; i <= LIMIT; i++)
    {
        for (j = 0; j <= endpoint; j+=2)
        {
            if (j == 0)
                p0 = grid[i][j];
            else
                p0 = (grid[i][j] + grid[i][j+1] ) / 2.0;
            p1 = grid[i][j+1];
            p2 = grid[i][j+2];
            if (j == endpoint)
                p3 = grid[i][j+3];
            else
                p3 = (grid[i][j+2] + grid[i][j+3]) / 2.0;

            t = (1.0/3.0);
            grid[i][j+1] = (1-t)*(1-t)*(1-t)*p0 + 3*t*(1-t)*(1-t)*p1 +
                3*t*t*(1-t)*p2 + t*t*t*p3;
            t = (2.0/3.0);
            grid[i][j+2] = (1-t)*(1-t)*(1-t)*p0 + 3*t*(1-t)*(1-t)*p1 +
                3*t*t*(1-t)*p2 + t*t*t*p3;
        }
    }

//  Do the columns
    for (j = 0; j <= LIMIT; j++)
    {
        for (i = 0; i <= endpoint; i+=2)
        {
            if (j == 0)
                p0 = grid[i][j];
            else
                p0 = (grid[i][j] + grid[i+1][j] ) / 2.0;
            p1 = grid[i+1][j];
            p2 = grid[i+2][j];
            if (j == endpoint)
                p3 = grid[i+3][j];
            else
                p3 = (grid[i+2][j] + grid[i+3][j]) / 2.0;

            t = (1.0/3.0);
            grid[i+1][j] = (1-t)*(1-t)*(1-t)*p0 + 3*t*(1-t)*(1-t)*p1 +
                3*t*t*(1-t)*p2 + t*t*t*p3;
            t = (2.0/3.0);
            grid[i+2][j] = (1-t)*(1-t)*(1-t)*p0 + 3*t*(1-t)*(1-t)*p1 +
                3*t*t*(1-t)*p2 + t*t*t*p3;
        }
    }
}



// This routine smoothes the landscape with a Lagrange interpolating
// polynomial
void smooth_lagrange()
{
    int i, j, endpoint;

    endpoint = LIMIT-3;
    
    // Do the rows
    // Since we will be copying grid to interp and adding more data, we will
    // work with every other row of interp.
    for (i = 0; i <= LIMIT; i++)
    {
        for (j = 0; j <= endpoint; j+=3)
        {
            lagrange_rows(i, j);
        }

        // Do the last few elements in the row
        lagrange_rows(i, endpoint);
    }


    endpoint = 2*LIMIT - 6;

    // Do the columnz, boyzzzz
    for (j = 0; j <= 2*LIMIT; j++)
    {
        for (i = 0; i <= endpoint; i+=6)
        {
            lagrange_cols(i, j);
        }
        // Do the last few elements in the column
        lagrange_cols(endpoint, j);
    }
}


// Note that this may not be the most efficient way to do these computations.
void lagrange_rows(int i, int j)
{
    int x0, x1, x2, x3;   // the x-values (grid location)
    GLdouble a, b, c, d;  // the y-values (grid height)
    int x;                // The interpolation point
    int k;

    x0 = 2*j;     x1 = 2*(j+1);
    x2 = 2*(j+2); x3 = 2*(j+3);
    // Change grid to patchgrid for some goodness
    // Set the base points (0, 2, 4, 6)
    a = interp[2*i][x0] = grid[i][j];
    b = interp[2*i][x1] = grid[i][j+1];
    c = interp[2*i][x2] = grid[i][j+2];
    d = interp[2*i][x3] = grid[i][j+3];

    // Set interpolated points (1, 3, 5) (9, 11, 13)
    for (k = 0; k < 3; k++)
    {
        x = 2 * (j+k) + 1;   // Like 2*x_k + 1
        interp[2*i][x] =  -a*(x-x1)*(x-x2)*(x-x3)/48
                         + b*(x-x0)*(x-x2)*(x-x3)/16
                         - c*(x-x0)*(x-x1)*(x-x3)/16
                         + d*(x-x0)*(x-x1)*(x-x2)/48;
    }
}


void lagrange_cols(int i, int j)
{
    int x0, x1, x2, x3;   // the x-values (grid location)
    GLdouble a, b, c, d;  // the y-values (grid height)
    int x;                // The interpolation point
    int k;

    x0 = i;   x1 = i+2;
    x2 = i+4; x3 = i+6;
    
    // The base points were copied above, just calculate the
    // interpolated points.
    a = interp[x0][j];
    b = interp[x1][j];
    c = interp[x2][j];
    d = interp[x3][j];

    // Set interpolated points (1, 3, 5) (9, 11, 13)
    for (k = 0; k < 3; k++)
    {
        x = i + 2*k + 1;
        interp[x][j] =  -a*(x-x1)*(x-x2)*(x-x3)/48
                       + b*(x-x0)*(x-x2)*(x-x3)/16
                       - c*(x-x0)*(x-x1)*(x-x3)/16
                       + d*(x-x0)*(x-x1)*(x-x2)/48;
    }
}



void smooth_patch()
{
    int i, j, endpoint;

    for (i = 0; i <= LIMIT; i++)
        memcpy(patchgrid[i], grid[i], (LIMIT+1)*sizeof(GLdouble));

    endpoint = LIMIT - 3;

    for (i = 0; i <= endpoint + 1; i+=2)
    {
        if (i > endpoint) i = endpoint;

        for (j = 0; j <= endpoint; j+=2)
        {
            compute_points(i,j);
            compute_points(i, endpoint);
        }
    }
}


//void compute_points(int i, int j)
//{
//    GLdouble p[4][4];
//    int endpoint = LIMIT - 3;
//    
//    // Compute left side of grid
//    if (j == 0)
//    {
//        p[0][0] = grid[i][j];
//        p[1][0] = grid[i+1][j];
//        p[2][0] = grid[i+2][j];
//    }
//    else
//    {
//        p[0][0] = (grid[i][j] + grid[i][j+1]) / 2.0;
//        p[1][0] = (grid[i+1][j] + grid[i+1][j+1]) / 2.0;
//        p[2][0] = (grid[i+2][j] + grid[i+2][j+1]) / 2.0;
//    }
//
//    // Compute middle points
//    p[0][1] = grid[i][j+1];   p[0][2] = grid[i][j+2];
//    p[1][1] = grid[i+1][j+1]; p[1][2] = grid[i+1][j+2];
//    p[2][1] = grid[i+2][j+1]; p[2][2] = grid[i+2][j+2];
//
//    // Compute right side of grid
//    if (j == endpoint)
//    {
//        p[0][3] = grid[i][j+3];
//        p[1][3] = grid[i+1][j+3];
//        p[2][3] = grid[i+2][j+3];
//    }
//    else
//    {
//        p[0][3] = (grid[i][j+2] + grid[i][j+3]) / 2.0;
//        p[1][3] = (grid[i+1][j+2] + grid[i+1][j+3]) / 2.0;
//        p[2][3] = (grid[i+2][j+2] + grid[i+2][j+3]) / 2.0;
//    }
//    
//    // Compute bottom of grid
//    if (i == endpoint)
//    {
//        p[3][0] = grid[i+3][j];
//        p[3][1] = grid[i+3][j+1];
//        p[3][2] = grid[i+3][j+2];
//        p[3][3] = grid[i+3][j+3];
//    }
//    else
//    {
//        p[3][0] = (grid[i+2][j] + grid[i+3][j]) / 2.0;
//        p[3][1] = (grid[i+2][j+1] + grid[i+3][j+1]) / 2.0;
//        p[3][2] = (grid[i+2][j+2] + grid[i+3][j+2]) / 2.0;
//        p[3][3] = (grid[i+2][j+3] + grid[i+3][j+2] + 2*grid[i+3][j+3]) / 4.0;
//    }
//
//    patchgrid[i+1][j+1] = patch_eval(p, 1.0/3.0, 1.0/3.0);
//    patchgrid[i+1][j+2] = patch_eval(p, 1.0/3.0, 2.0/3.0);
//    patchgrid[i+2][j+1] = patch_eval(p, 2.0/3.0, 1.0/3.0);
//    patchgrid[i+2][j+2] = patch_eval(p, 2.0/3.0, 2.0/3.0);
//
//}

#define max(a,b)            (((a) > (b)) ? (a) : (b))
#define min(a,b)            (((a) < (b)) ? (a) : (b))

GLdouble average_points( int i, int j )
{
    int left, top, right, bottom;

    left = max( i-1, 0 );
    top  = max( j-1, 0 );
    right = min( i+1, LIMIT );
    bottom = min( j+1, LIMIT );

    return ( grid[i][j] + grid[left][j] + grid[right][j] + grid[i][top] + grid[i][bottom] ) / 5;
}

void compute_points(int i, int j)
{
    GLdouble p[4][4];
    int endpoint = LIMIT - 3;
    
    p[0][0] = average_points( i, j );
    p[0][3] = average_points( i, j+3 );
    p[3][0] = average_points( i+3, j );
    p[3][3] = average_points( i+3, j+3 );


    // left side
    p[1][0] = grid[i+1][j];
    p[2][0] = grid[i+2][j];
    // right side
    p[1][3] = grid[i+1][j+3];
    p[2][3] = grid[i+2][j+3];

    // Compute middle points
    p[0][1] = grid[i][j+1];   p[0][2] = grid[i][j+2];
    p[1][1] = grid[i+1][j+1]; p[1][2] = grid[i+1][j+2];
    p[2][1] = grid[i+2][j+1]; p[2][2] = grid[i+2][j+2];
    p[3][1] = grid[i+3][j+1]; p[3][2] = grid[i+3][j+2];



    
    patchgrid[i+1][j+1] = patch_eval(p, 1.0/3.0, 1.0/3.0);
    patchgrid[i+1][j+2] = patch_eval(p, 1.0/3.0, 2.0/3.0);
    patchgrid[i+2][j+1] = patch_eval(p, 2.0/3.0, 1.0/3.0);
    patchgrid[i+2][j+2] = patch_eval(p, 2.0/3.0, 2.0/3.0);

}

double patch_eval(double p[4][4], double u, double v)
{
    double B0u = (1-u)*(1-u)*(1-u);
    double B1u = 3*u*(1-u)*(1-u);
    double B2u = 3*u*u*(1-u);
    double B3u = u*u*u;
    double B0v = (1-v)*(1-v)*(1-v);
    double B1v = 3*v*(1-v)*(1-v);
    double B2v = 3*v*v*(1-v);
    double B3v = v*v*v;

    return
    p[0][0]*B0u*B0v + p[0][1]*B0u*B1v + p[0][2]*B0u*B2v + p[0][3]*B0u*B3v
  + p[1][0]*B1u*B0v + p[1][1]*B1u*B1v + p[1][2]*B1u*B2v + p[1][3]*B1u*B3v
  + p[2][0]*B2u*B0v + p[2][1]*B2u*B1v + p[2][2]*B2u*B2v + p[2][3]*B2u*B3v
  + p[3][0]*B3u*B0v + p[3][1]*B3u*B1v + p[3][2]*B3u*B2v + p[3][3]*B3u*B3v;

}